
function [x,flag] = ubils_admm(B,y,l,u,cp,maxtime,noise_var,lower_bound_levels)
% 
% x = ubils_admm(B,y,l,u) uses ADMM to aid the search process to produces 
%     the optimal solution or a suboptimal solution to the underdetermined 
%     box-constrained integer least square problem min_{l<=x<=u}||y-Bx||.
%     B - m-by-n real matrix with rank(B) < n
%     y - m-dimensional real vector
%     l - n-dimensional integer vector, lower bound 
%     u - n-dimensional integer vector, upper bound, l < u
%     x - n-dimensional integer vector (in double precision), the optimal
%         solution 
% 
% x = ubils_admm(B,y,l,u,cp) provides a column permutation option to choose
%     cp = 0 / [] (default), performs simple column permutations;
%     cp = 1, performs sophasitated column permutations, preferable for 
%             large resdiual cases.
%  
% x = ubils_admm(B,y,l,u,cp,maxtime) specifies the maximum running time in 
%     seconds, and x may be optimal or suboptimal.  If maxtime is [] then 
%     ubils_admm uses the default, Inf, and x is the optimal solution.
%
% x = ubils_admm(B,y,l,u,cp,maxtime,noise_var) specifies variance of 
%     the elements of teh noise v in the model y = Ax + v. If noise_var 
%     is [] then ubils_admm uses the default, 0.1.
%
% x = ubils_admm(B,y,l,u,cp,maxtime,noise_var,lower_bound_levels) specifies
%     the number of lower bound levels to be found by ADMM to reduce the 
%     search region. You may set lower_bound_levels to be a value between 1
%     and n-m. The default is round((n-m)/2). 
%
%  [x,flag] = ubils_admm(B,y,l,u,...) also returns a flag:
%     flag = 0, x is the optimal solution
%     flag = 1, x is a suboptimal solution 

% Subfunctions: ubils_reduction, ubils_search

% Main References:
% [1] X.-W. Chang and X. Yang. An efficient tree search
%     decoder with column reordering for underdetermined MIMO systems. 
%     Proceedings of IEEE GLOBECOM 2007, 5 pages.
% [2] Tianchi Ma. An ADMM Method for Underdetermined Box-constrained Integer
%     Least Quuares Problems. M.Sc. Thesis, School of Computer Science, 
%     McGill University, May 2021.

% Authors: Xiao-Wen Chang, www.cs.mcgill.ca/~chang 
%           Tianchi Ma, Zhilong Chen
% Copyright (c) 2022. Scientific Computing Lab, McGill University.
% Last revision: December 2022

[m,n] = size(B);

% Check the input arguments
if nargin > 8
    error('Too many input arguments');
end

if nargin < 4
    error('Not enought input arguments');
end

switch nargin
    case 4 
        cp = 0;
        maxtime = Inf;
        noise_var = 0.1;
        lower_bound_levels = round((n-m)/2);
    case 5 
        if isempty(cp)
           cp = 0;
        end
        if cp ~= 0 &&  cp ~= 1
            error('cp should be [], 0 or 1');
        end
        maxtime = Inf;
        noise_var = 0.1;
        lower_bound_levels = round((n-m)/2);
    case 6
        if isempty(cp)
           cp = 0;
        end
        if cp ~= 0 &&  cp ~= 1
            error('cp should be [], 0 or 1');
        end
        if maxtime <= 0
           error('Maxtime has to be positive');
        end
        noise_var = 0.1;
        lower_bound_levels = round((n-m)/2);
     case 7
        if isempty(cp)
           cp = 0;
        end
        if cp ~= 0 &&  cp ~= 1
            error('cp should be [], 0 or 1');
        end
        if maxtime <= 0
           error('Maxtime has to be positive');
        end
        if noise_var < 0
           error('Noise variance has to be non-negative');
        end
        lower_bound_levels = round((n-m)/2);
      case 8
        if isempty(cp)
           cp = 0;
        end
        if cp ~= 0 &&  cp ~= 1
            error('cp should be [], 0 or 1');
        end
        if maxtime <= 0
           error('Maxtime has to be positive');
        end
        if noise_var < 0
           error('Noise variance has to be non-negative');
        end
        if lower_bound_levels < 0
           error('lower_bound_levels has to be no samller than 1');
        end
end

if m ~= size(y,1) || size(y,2) ~= 1 || ... 
       n ~= size(l,1) || size(l,2) ~= 1 || ...
       n ~= size(u,1) || size(u,2) ~= 1     % Input error
    error('Input arguments have a matrix dimension error')
end

l = ceil(l); u = floor(u);  % To make it work with real bounds
for i = 1 : n
    if l(i) >= u(i) || l(i) == -Inf || u(i) == Inf
        error('Invalid lower bound or upper bound');
    end
end


% If B is a zero matrix, one solution is obtained
if B == 0
    x = l;
    flag = 0;
    warning('The input matrix is zero and one solution is produced')
    return
end

tic

% Reduction - reduce the problem to the trapezoid form
[R,y,l,u,p,z0,beta] = ubils_reduction_initial(B,y,l,u,cp,noise_var);

duetime = maxtime - toc;

% Search 
if beta > n*eps*norm(abs(y)+abs(R)*abs(z0))
    [z,flag] = ubils_search_lowerbound(R,y,l,u,z0,beta,duetime,noise_var,lower_bound_levels);  
else
    z = z0; % z0 is the solution of the reduced problem
    flag = 0;
end

% Reorder z to obtain the solution to the original problem
x = zeros(n,1);
for i = 1 : n
    x(p(i)) = z(i);
end 

